"""
Test case with additional test methods used by some Annalist tests
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import re

import logging
log = logging.getLogger(__name__)

from django.test import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions

from annalist.models.entitytypeinfo    import EntityTypeInfo

from annalist.views.fields.bound_field import bound_field

class AnnalistTestCase(TestCase):
    """
    Additonal test methods for Annalist test cases
    """

    def check_entity_values(self, type_id, entity_id, check_values=None):
        "Helper function checks content of entity record"
        typeinfo = EntityTypeInfo(self.testsite, self.testcoll, type_id)
        self.assertTrue(typeinfo.entity_exists(entity_id))
        t = typeinfo.get_entity(entity_id)
        self.assertEqual(t.get_id(), entity_id)
        self.assertEqual(t.get_type_id(), type_id)
        self.assertDictionaryMatch(t.get_values(), check_values)
        return t

    def assertEqualPrefix(self, actual, expect, prefix=""):
        self.assertEqual(actual, expect, msg="%s: actual %r, expected %r"%(prefix, actual, expect))
        return

    def assertMatch(self, string, pattern, msg=None):
        """
        Throw an exception if the regular expresson pattern is matched
        """
        m = re.search(pattern, string)
        if not m or not m.group(0):
            raise self.failureException(
                (msg or "'%s' does not match /%s/"%(string, pattern)))
        return

    def assertKeysMatch(self, actual_dict, expect_dict):
        """
        Check that the expect_dict keys are the same as those in actual_dict.

        Use with assertDictionaryMatch to ensure that two dictionaries are the same,
        not just that all the expected entries are present in the actual dictionary.
        """
        self.assertEqual(set(actual_dict.keys()), set(expect_dict.keys()))
        return

    def assertDictionaryMatch(self, actual_dict, expect_dict, prefix=""):
        """
        Check that the expect_dictr values are all present in actual_dict.

        If a dictionary element contains a list, the listed values are assumed to
        to be dictionaries which are matched recursively. (This logic is used when 
        checking sub-contexts used to render data-defined forms.)
        """
        # log.info("\n***********\nexpect_dict %r"%(expect_dict))
        # log.info("\n-----------\nactual_dict %r"%(actual_dict))
        for k in expect_dict:
            self.assertTrue(k in actual_dict, prefix+"Expected key %s not found in actual"%(k))
            if isinstance(expect_dict[k],list):
                if isinstance(actual_dict[k],list):
                    # log.info("dict match: prefix: %s, key: %s, list_len %d"%(prefix, k, len(expect_dict[k])))
                    if k == "@type":
                        self.assertEqual(
                            set(actual_dict[k]), set(expect_dict[k]),
                            "Key %s: %r != %r"%(k, set(actual_dict[k]), set(expect_dict[k]))
                            )
                    else:
                        for i in range(len(expect_dict[k])):
                            # if i >= len(actual_dict[k]):
                            #     log.info("\n***********\nexpect_dict %r"%(expect_dict))
                            #     log.info("\n-----------\nactual_dict %r"%(actual_dict))
                            #     log.info("\n***********")
                            self.assertTrue(i < len(actual_dict[k]), prefix+"Actual dict key %s has no element %d"%(k,i))
                            if ( isinstance(actual_dict[k][i], (dict, bound_field)) and 
                                 isinstance(expect_dict[k][i], dict)
                               ):
                                self.assertDictionaryMatch(
                                    actual_dict[k][i], expect_dict[k][i], 
                                    prefix=prefix+"Key %s[%d]: "%(k,i)
                                    )
                            else:
                                self.assertEqual(
                                    actual_dict[k][i], expect_dict[k][i],
                                    prefix+"Key %s[%d]: %r != %r"%(k, i, actual_dict[k], expect_dict[k])
                                    )
                else:
                    self.fail("Key %s expected list, got %r instead of %r"%(k, actual_dict[k], expect_dict[k]))
            else:
                # if actual_dict[k] != expect_dict[k]:
                #     log.info("\n***********\nexpect_dict %r"%(expect_dict))
                #     log.info("\n-----------\nactual_dict %r"%(actual_dict))
                #     log.info("\n***********")
                # log.info("dict match: prefix: %s, key: %s, actual %s, expected: %s"%(prefix, k, actual_dict[k], expect_dict[k]))
                self.assertEqual(actual_dict[k], expect_dict[k], 
                    prefix+"Key %s: actual '%s' expected '%s'"%(k, actual_dict[k], expect_dict[k]))
        # log.info("\n****** matched")
        return

# End.



"""
Tests for EntityData subtype selection in list views and select dropdowns
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.test.client             import Client

from annalist.models.site           import Site
from annalist.models.collection     import Collection
from annalist.models.recordtype     import RecordType
from annalist.models.recordview     import RecordView
from annalist.models.recordfield    import RecordField
from annalist.models.recordtypedata import RecordTypeData
from annalist.models.entitydata     import EntityData

from annalist.views.form_utils.fieldchoice  import FieldChoice

from tests                          import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from tests                          import init_annalist_test_site, resetSitedata
from AnnalistTestCase               import AnnalistTestCase
from entity_testutils               import (
    collection_create_values,
    create_test_user,
    context_list_entities,
    )
from entity_testtypedata            import (
    recordtype_create_values
    )
from entity_testentitydata          import (
    entity_url, entitydata_edit_url, entitydata_delete_confirm_url,
    entitydata_list_type_url, entitydata_list_all_url,
    entitydata_value_keys, entitydata_create_values, entitydata_values, 
    )

#   -----------------------------------------------------------------------------
#
#   Subtype selection tests
#
#   -----------------------------------------------------------------------------

class SubtypeSelectionTest(AnnalistTestCase):
    """
    Tests for 
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite  = Site(TestBaseUri, TestBaseDir)
        self.testcoll  = Collection.create(
            self.testsite, "testcoll", collection_create_values("testcoll")
            )
        # Create test types
        self.testtypes = RecordType.create(
            self.testcoll, "testtypes", 
            recordtype_create_values(
                coll_id="testcoll", type_id="testtypes", type_uri="test:testtypes",
                supertype_uris=[]
                )
            )
        self.testtype1 = RecordType.create(
            self.testcoll, "testtype1",
            recordtype_create_values(
                coll_id="testcoll", type_id="testtype1", type_uri="test:testtype1", 
                supertype_uris=["test:testtypes"]
                )
            )
        self.testtype2 = RecordType.create(
            self.testcoll, "testtype2",
            recordtype_create_values(
                coll_id="testcoll", type_id="testtype2", type_uri="test:testtype2", 
                supertype_uris=["test:testtypes"]
                )
            )
        self.ref_type  = RecordType.create(
            self.testcoll, "ref_type", 
            recordtype_create_values(
                coll_id="testcoll", type_id="ref_type", type_uri="test:ref_type",
                supertype_uris=[]
                )
            )
        # Create test type data parents
        self.testdatas = RecordTypeData.create(self.testcoll, "testtypes", {})
        self.testdata1 = RecordTypeData.create(self.testcoll, "testtype1", {})
        self.testdata2 = RecordTypeData.create(self.testcoll, "testtype2", {})
        self.ref_data  = RecordTypeData.create(self.testcoll, "ref_type",  {})
        # Create test type data
        es = EntityData.create(self.testdatas, "entitys", 
            entitydata_create_values(
                "entitys", type_id="testtypes", extra_fields={"test:turi": "test:testtypes"} 
                )
            )
        e1 = EntityData.create(self.testdata1, "entity1", 
            entitydata_create_values(
                "entity1", type_id="testtype1", extra_fields={"test:turi": "test:testtype1"} 
                )
            )
        e2 = EntityData.create(self.testdata2, "entity2", 
            entitydata_create_values(
                "entity2", type_id="testtype2", extra_fields={"test:turi": "test:testtype2"} 
                )
            )
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        # resetSitedata(scope="collections")
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    #   -----------------------------------------------------------------------------
    #   Helpers
    #   -----------------------------------------------------------------------------

    def _check_entity_list(self, type_id, expect_entities):
        u = entitydata_list_type_url("testcoll", type_id)
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<title>Collection testcoll</title>")
        self.assertContains(r, "<h3>List entities</h3>", html=True)
        # Test context
        self.assertEqual(r.context['coll_id'],  "testcoll")
        self.assertEqual(r.context['type_id'],  type_id)
        entities = context_list_entities(r.context)
        self.assertEqual(len(entities), len(expect_entities))
        for eid in range(len(expect_entities)):
            self.assertEqual(entities[eid]['entity_type_id'], expect_entities[eid][0])
            self.assertEqual(entities[eid]['entity_id'], expect_entities[eid][1])
        return

    def _create_ref_type_view(self, view_id="Test_ref_type_view", record_type="test:testtypes"):
        ref_type_view = RecordView.create(self.testcoll, view_id,
            { 'annal:type':         "annal:View"
            , 'annal:uri':          "test:"+view_id
            , 'rdfs:label':         "Test view label"
            , 'rdfs:comment':       "Test view comment"
            , 'annal:record_type':  record_type
            , 'annal:add_field':    True
            , 'annal:view_fields':
              [ { 'annal:field_id':         "_field/Entity_id"
                , 'annal:field_placement':  "small:0,12;medium:0,6"
                }
              , { 'annal:field_id':         "_field/Test_ref_type_field"
                , 'annal:field_placement':  "small:0,12;medium:0,6"
                }
              ]
            })
        self.assertTrue(ref_type_view is not None)
        return ref_type_view

    def _create_ref_type_field(self, entity_type="test:testtypes"):
        ref_type_field = RecordField.create(self.testcoll, "Test_ref_type_field",
            { "annal:type":                     "annal:Field"
            , "rdfs:label":                     "Type reference"
            , "rdfs:comment":                   "Type reference field comment"
            , "annal:field_render_type":        "Enum_render_type/Enum_choice"
            , "annal:field_value_mode":         "Enum_value_mode/Value_direct"
            , "annal:field_entity_type":        "test:ref_type"
            , "annal:placeholder":              "(ref type field)"
            , "annal:property_uri":             "test:ref_type"
            , "annal:field_placement":          "small:0,12;medium:0,6"
            , "annal:field_ref_type":           "_type/testtypes"
            , "annal:field_ref_restriction":    "[test:turi] subtype %s"%entity_type
            , "annal:field_entity_type":        entity_type
            })
        self.assertTrue(ref_type_field is not None)
        return ref_type_field

    #   -----------------------------------------------------------------------------
    #   Subtype rendering tests
    #   -----------------------------------------------------------------------------

    def test_list_type1(self):
        expect_entities = (
            [ ("testtype1", "entity1") 
            ])
        self._check_entity_list("testtype1", expect_entities)
        return

    def test_list_type2(self):
        expect_entities = (
            [ ("testtype2", "entity2") 
            ])
        self._check_entity_list("testtype2", expect_entities)
        return

    def test_list_types(self):
        expect_entities = (
            [ ("testtype1", "entity1") 
            , ("testtype2", "entity2") 
            , ("testtypes", "entitys") 
            ])
        self._check_entity_list("testtypes", expect_entities)
        return

    def test_select_subtypes(self):
        ref_view  = self._create_ref_type_view()
        ref_field = self._create_ref_type_field()
        u = entitydata_edit_url(
            "new", "testcoll", "ref_type", view_id="Test_ref_type_view"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Check render context
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "ref_type")
        # Fields
        self.assertEqual(len(r.context['fields']), 2)
        self.assertEqual(r.context['fields'][0]['field_id'], "Entity_id")
        self.assertEqual(r.context['fields'][1]['field_id'], "Test_ref_type_field")
        baselabel = "Entity testcoll/"
        baseuri   = TestBasePath+"/c/testcoll/d/%s/"
        ref_options = (
            [ FieldChoice(opt, label=baselabel+opt, link=baseuri%opt)
              for opt in ['testtype1/entity1', 'testtype2/entity2', 'testtypes/entitys']
            ])
        self.assertEqual(r.context['fields'][1]['options'], ref_options)
        return

    def test_select_notype_fields(self):
        ref_view  = self._create_ref_type_view(view_id="Test_ref_notype_view", record_type="test:notypes")
        ref_field = self._create_ref_type_field(entity_type="test:sometypes")
        u = entitydata_edit_url(
            "new", "testcoll", "ref_type", view_id="Test_ref_notype_view"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Check render context
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "ref_type")
        # Fields
        self.assertEqual(len(r.context['fields']), 2)
        self.assertEqual(r.context['fields'][0]['field_id'], "Entity_id")
        self.assertEqual(r.context['fields'][1]['field_id'], "Test_ref_type_field")
        baselabel = "Entity testcoll/"
        baseuri   = TestBasePath+"/c/testcoll/d/%s/"
        ref_options = []
        self.assertEqual(r.context['fields'][1]['options'], ref_options)
        return

    def test_select_subtype1_fields(self):
        ref_view  = self._create_ref_type_view(view_id="Test_ref_type1_view", record_type="test:testtypes")
        ref_field = self._create_ref_type_field(entity_type="test:testtype1")
        u = entitydata_edit_url(
            "new", "testcoll", "ref_type", view_id="Test_ref_type1_view"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Check render context
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "ref_type")
        # Fields
        self.assertEqual(len(r.context['fields']), 2)
        self.assertEqual(r.context['fields'][0]['field_id'], "Entity_id")
        self.assertEqual(r.context['fields'][1]['field_id'], "Test_ref_type_field")
        baselabel = "Entity testcoll/"
        baseuri   = TestBasePath+"/c/testcoll/d/%s/"
        ref_options = (
            [ FieldChoice(opt, label=baselabel+opt, link=baseuri%opt)
              for opt in ['testtype1/entity1']
            ])
        self.assertEqual(r.context['fields'][1]['options'], ref_options)
        return

    def _no_test_test_select_view_subtype_fields(self):
        # @@INCOMPLETE:  this was an attempt to check field selction from the view context,
        #                but abandoned to to complexity of reconstructing bound fields.
        #                NOTE: sitedata tests use Beautifulsoup representation of rendered
        #                page for these tests
        ref_view  = self._create_ref_type_view()
        ref_field = self._create_ref_type_field()
        u = entitydata_edit_url(
            "edit", "testcoll", "_view", "Test_ref_type_view", view_id="View_view"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Check render context
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "_view")
        # Fields
        self.assertEqual(len(r.context['fields']), 6)
        self.assertEqual(r.context['fields'][0]['field_id'], "View_id")
        self.assertEqual(r.context['fields'][1]['field_id'], "View_label")
        self.assertEqual(r.context['fields'][2]['field_id'], "View_comment")
        self.assertEqual(r.context['fields'][3]['field_id'], "View_target_type")
        self.assertEqual(r.context['fields'][4]['field_id'], "View_edit_view")
        self.assertEqual(r.context['fields'][5]['field_id'], "View_fields")
        print repr(r.context['fields'][5])
        baselabel = "Entity testcoll/"
        baseuri   = TestBasePath+"/c/testcoll/d/%s/"
        ref_options = (
            [ FieldChoice(opt, label=baselabel+opt, link=baseuri%opt)
              for opt in ['testtype1/entity1', 'testtype2/entity2', 'testtypes/entitys']
            ])
        self.assertEqual(r.context['fields'][5]['options'], ref_options)
        return

# End.



"""
Tests for boolean value rendering as checkbox
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import sys
import os
import unittest
import re
from collections import OrderedDict

import logging
log = logging.getLogger(__name__)

# from django.conf                                import settings
# from django.test                                import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions
# from django.template                            import Context, Template, loader

from annalist.views.fields.render_bool_checkbox import (
    get_bool_checkbox_renderer, 
    BoolCheckboxValueMapper
    )

from annalist.tests.field_rendering_support     import FieldRendererTestSupport

class BooleanCheckboxRenderingTest(FieldRendererTestSupport):

    def setUp(self):
        return

    def tearDown(self):
        return

    def test_RenderBoolCheckboxValue(self):

        def expect_render(val):
            if isinstance(val, str):
                valtext = val
                valbool = val.lower() in ["true", "yes"]
            else:
                valtext = "Yes" if val else "No"
                valbool = val
            checked     = ''' checked="checked"''' if valbool else ''''''
            render_view = '''<span>%s</span>'''%valtext
            render_edit = (
                '''<input type="checkbox" '''+
                       '''name="repeat_prefix_test_field" '''+
                       ('''value="%s"%s'''%(valtext,checked)) +
                       ''' />'''+
                  ''' <span class="value-placeholder">(test placeholder)</span>'''
                )
            return {'view': render_view, 'edit': render_edit}

        test_value_context_renders = (
            [ (self._make_test_context(None),    expect_render(False))
            , (self._make_test_context(False),   expect_render(False))
            , (self._make_test_context(True),    expect_render(True))
            , (self._make_test_context("False"), expect_render("False"))
            , (self._make_test_context("True"),  expect_render("True"))
            , (self._make_test_context("Yes"),   expect_render("Yes"))
            , (self._make_test_context("No"),    expect_render("No"))
            , (self._make_test_context(u"yes"),  expect_render("yes"))
            , (self._make_test_context(u"no"),   expect_render("no"))
            ])
        renderer = get_bool_checkbox_renderer()

        for render_context, expect_render in test_value_context_renders:
            # print repr(render_context['field']['field_value'])
            self._check_value_renderer_results(
                renderer,
                context=render_context,
                expect_rendered_view=expect_render['view'],
                expect_rendered_edit=expect_render['edit']
                )
        return

    def test_DecodeBoolCheckboxValue(self):
        # Any text other than None or empty string indicates the box is checked, returns True
        test_decode_values = (
            { None:         False
            , "":           False
            , "checked":    True
            , "No":         True
            , "Yes":        True
            , "False":      True
            , "True":       True
            , "false":      True
            , "true":       True
            })
        for valtext, expect_valdata in test_decode_values.items():
            valdata = BoolCheckboxValueMapper.decode(valtext)
            self.assertEqual(
                valdata, expect_valdata, 
                "Value decode(%s) = %r, expect %r"%(valtext, valdata, expect_valdata)
                )
        return

# End.

if __name__ == "__main__":
    # import django
    # django.setup()  # Needed for template loader
    # Runtests in this module
    # runner = unittest.TextTestRunner(verbosity=2)
    # tests = unittest.TestSuite()
    # tests  = getSuite(select=sel)
    # if tests: runner.run(tests)
    unittest.main()



"""
This module contains (and isolates) logic used to find entities based on entity type,
list selection criteria and search terms.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import re
from pyparsing import Word, QuotedString, Literal, Group, Empty, StringEnd, ParseException
from pyparsing import alphas, alphanums

from annalist.util                  import extract_entity_id

from annalist.models.recordtypedata import RecordTypeData
from annalist.models.entitytypeinfo import EntityTypeInfo, get_built_in_type_ids

#   -------------------------------------------------------------------
#   Auxilliary functions
#   -------------------------------------------------------------------

def order_entity_key(entity):
    """
    Function returns sort key for ordering entities by type and entity id

    Use with `sorted`, thus:

        sorted(entities, order_entity_key)
    """
    type_id   = entity.get_type_id() 
    entity_id = entity.get_id()
    key = ( 0 if type_id.startswith('_')   else 1, type_id, 
            0 if entity_id.startswith('_') else 1, entity_id
          )
    # log.info(key)
    return key

#   -------------------------------------------------------------------
#   EntityFinder
#   -------------------------------------------------------------------

class EntityFinder(object):
    """
    Logic for enumerating entities matching a supplied type, selector and/or search string.
    """

    def __init__(self, coll, selector=None):
        """
        Initialize entity finder for collection and selector.
        """
        super(EntityFinder, self).__init__()
        self._coll     = coll
        self._site     = coll.get_site()
        self._selector = EntitySelector(selector, FieldComparison(coll))
        # self._subtypes = None
        return

    def get_collection_type_ids(self):
        """
        Returns iterator over possible type ids in current collection.

        Each type is returned as a candidate type identifier string
        """
        for t in get_built_in_type_ids():
            yield t
        for t in self._coll._children(RecordTypeData):
            yield t
        return

    def get_collection_subtypes(self, type_id):
        """
        Returns a iterator of `entitytypeinfo` objects for all subtypes
        of the supplied type in the current collection, including the 
        identified type itself.
        """
        supertypeinfo = EntityTypeInfo(self._site, self._coll, type_id)
        supertypeuri  = supertypeinfo.get_type_uri()
        return self.get_collection_uri_subtypes(supertypeuri)

    def get_collection_uri_subtypes(self, type_uri):
        """
        Returns a iterator of `entitytypeinfo` objects for all subtypes
        of the supplied type in the current collection, including the 
        identified type itself.
        """
        if type_uri is not None:
            for tid in self.get_collection_type_ids():
                tinfo = EntityTypeInfo(self._site, self._coll, tid)
                if tinfo and (type_uri in tinfo.get_all_type_uris()):
                    # log.info(
                    #     "supertype %s, yield %s: %s"%
                    #     (type_uri, tinfo.get_type_id(), tinfo.get_type_uri())
                    #     )
                    yield tinfo
        return 

    def get_type_entities(self, type_id, user_permissions, scope):
        """
        Iterate over entities from collection matching the supplied type.

        'scope' is used to determine the extent of data top be included in the listing:
        a value of 'all' means that site-wide entyioties are icnluded in the listing.
        Otherwise only collection entities are included.        
        """
        entitytypeinfo = EntityTypeInfo(self._site, self._coll, type_id)
        include_sitedata = (scope == "all")
        for e in entitytypeinfo.enum_entities_with_inferred_values(
                user_permissions, usealtparent=include_sitedata
                ):
            yield e
        return

    def get_subtype_entities(self, type_id, user_permissions, scope):
        """
        Iterate over entities from collection that are of the indicated type
        or any of its subtypes.

        'scope' is used to determine the extent of data to be included in the listing:
        a value of 'all' means that site-wide entities are included in the listing.
        Otherwise only collection entities are included.        
        """
        include_sitedata = (scope == "all")
        for entitytypeinfo in self.get_collection_subtypes(type_id):
            for e in entitytypeinfo.enum_entities_with_inferred_values(
                    user_permissions, usealtparent=include_sitedata
                    ):
                # log.info(
                #     "get_subtype_entities type_id %s, yield %s/%s"%
                #     (type_id, e.get_type_id(), e.get_id())
                #     )
                yield e
        return

    def get_all_types_entities(self, types, user_permissions, scope):
        """
        Iterate mover all entities of all type ids from a supplied type iterator
        """
        assert user_permissions is not None
        for t in types:
            for e in self.get_type_entities(t, user_permissions, scope):
                yield e
        return

    def get_base_entities(self, type_id=None, user_permissions=None, scope=None):
        """
        Iterate over base entities from collection, matching the supplied type id if supplied.

        If a type_id is supplied, site data values are included.
        """
        if type_id:
            return self.get_subtype_entities(type_id, user_permissions, scope)
            # return self.get_type_entities(type_id, user_permissions, scope)
        else:
            return self.get_all_types_entities(
                self.get_collection_type_ids(), user_permissions, scope
                )
        return

    def search_entities(self, entities, search):
        """
        Iterate over entities from supplied iterator containing supplied search term.
        """
        for e in entities:
            if self.entity_contains(e, search):
                yield e
        return

    def get_entities(self, 
        user_permissions=None, type_id=None, scope=None, context=None, search=None
        ):
        """
        Get list of entities of the specified type, matching search term and visible to 
        supplied user permissions.
        """
        entities = self._selector.filter(
            self.get_base_entities(type_id, user_permissions, scope), context=context
            )
        if search:
            entities = self.search_entities(entities, search)
        return entities

    def get_entities_sorted(self, 
        user_permissions=None, type_id=None, scope=None, context={}, search=None
        ):
        """
        Get sorted list of entities of the specified type, matching search term and 
        visible to supplied user permissions.
        """
        entities = self.get_entities(
            user_permissions, type_id=type_id, scope=scope, 
            context=context, search=search
            )
        return sorted(entities, key=order_entity_key)

    @classmethod
    def entity_contains(cls, e, search):
        """
        Returns True if entity contains/matches search term, else False.
        Search term None (or blank) matches all entities.

        >>> e  = { 'p:a': '1', 'p:b': '2', 'p:c': '3' }
        >>> EntityFinder.entity_contains(e, "1")
        True
        >>> EntityFinder.entity_contains(e, "3")
        True
        >>> EntityFinder.entity_contains(e, "nothere")
        False
        """
        if search:
            for key in e:
                val = e[key]
                if isinstance(val, (str, unicode)):
                    if search in val:
                        return True
            return False
        return True

#   -------------------------------------------------------------------
#   EntitySelector
#   -------------------------------------------------------------------

class EntitySelector(object):
    """
    This class implements a selector filter.  It is initialized with a selector
    expression, and may be invoked as a filter applied to an entity generator,
    or as a predicate applied to a single entity.

    >>> e  = { 'p:a': '1', 'p:b': '2', 'p:c': '3', '@type': ["http://example.com/type", "foo:bar"] }
    >>> c  = { 'view': { 'v:a': '1', 'v:b': ['2', '3'] } }
    >>> f1 = "'1' == [p:a]"
    >>> f2 = "[p:a]=='2'"
    >>> f3 = ""
    >>> f4 = "'http://example.com/type' in [@type]"
    >>> f5 = "'foo:bar' in [@type]"
    >>> f6 = "'bar:foo' in [@type]"
    >>> f7 = "[p:a] in view[v:a]"
    >>> f8 = "[p:b] in view[v:b]"
    >>> f9 = "[p:a] in view[v:b]"
    >>> f10 = "[annal:field_entity_type] in view[annal:record_type]"
    >>> f11 = "foo:bar in [@type]"
    >>> f12 = "bar:foo in [@type]"
    >>> EntitySelector(f1).select_entity(e, c)
    True
    >>> EntitySelector(f2).select_entity(e, c)
    False
    >>> EntitySelector(f3).select_entity(e, c)
    True
    >>> EntitySelector(f4).select_entity(e, c)
    True
    >>> EntitySelector(f5).select_entity(e, c)
    True
    >>> EntitySelector(f6).select_entity(e, c)
    False
    >>> EntitySelector(f7).select_entity(e, c)
    True
    >>> EntitySelector(f8).select_entity(e, c)
    True
    >>> EntitySelector(f9).select_entity(e, c)
    False
    >>> EntitySelector(f10).select_entity(e, c)
    True
    >>> EntitySelector(f11).select_entity(e, c)
    True
    >>> EntitySelector(f12).select_entity(e, c)
    False
    """
    def __init__(self, selector, fieldcomp=None):
        self._fieldcomp = fieldcomp
        # Returns None if no filter is applied, otherwise a predcicate function
        self._selector  = self.compile_selector_filter(selector)
        return

    def filter(self, entities, context=None):
        """
        Iterate over selection of entities from supplied iterator, using the
        selection specification supplied to the constructor of the current object.

        entities    is an iterator over entities from which selection is made
        context     is a dictionary of context values that may be referenced by 
                    the selector in choosing entities to be returned.

        If no filtering is applied, the supplied iterator is returned as-is.
        """
        if self._selector:
            entities = self._filter(entities, context)
        return entities

    def _filter(self, entities, context):
        """
        Internal helper applies selector to entity iterator, returning a new iterator.
        """
        for e in entities:
            if self._selector(e, context):
                yield e
        return

    def select_entity(self, entity, context={}):
        """
        Apply selector to an entity, and returns True if the entity is selected
        """
        if self._selector:
            return self._selector(entity, context)
        return True

    @classmethod  #@@ @staticmethod, no cls?
    def parse_selector(cls, selector):
        """
        Parse a selector and return list of tokens

        Selector formats:
            ALL (or blank)              match any entity
            <val1> == <val2>            values are same
            <val1> in <val2>            second value is list containing 1st value, 
                                        or values are same, or val1 is None.
            <val1> <name> <val2>        invoke comparison method from supplied 
                                        FieldComparison object

            <val1> and <val2> may be:

            [<field-id>]                refers to field in entity under test
            <name>[<field-id>]          refers to field of context value, or None if the
                                        indicated context value or field is not defined.
            "<string>"                  literal string value.  Quotes within are escaped.

        <field_id> values are URIs or CURIEs, using characters defined by RFC3986,
        except "[" and "]"
        
        RFC3986:
           unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
           reserved      = gen-delims / sub-delims
           gen-delims    = ":" / "/" / "?" / "#" / "[" / "]" / "@"
           sub-delims    = "!" / "$" / "&" / "'" / "(" / ")"
                         / "*" / "+" / "," / ";" / "="

        Parser uses pyparsing combinators (cf. http://pyparsing.wikispaces.com).
        """
        def get_value(val_list):
            if len(val_list) == 1:
                return { 'type': 'literal', 'name': None,        'field_id': None,        'value': val_list[0] }
            elif val_list[0] == '[':
                return { 'type': 'entity',  'name': None,        'field_id': val_list[1], 'value': None }
            elif val_list[1] == '[':
                return { 'type': 'context', 'name': val_list[0], 'field_id': val_list[2], 'value': None }
            else:
                return { 'type': 'unknown', 'name': None,        'field_id': None,        'value': None }
        p_name     = Word(alphas+"_", alphanums+"_")
        p_id       = Word(alphas+"_@", alphanums+"_-.~:/?#@!$&'()*+,;=)")
        p_val      = ( Group( Literal("[") + p_id + Literal("]") )
                     | Group( p_name + Literal("[") + p_id + Literal("]") )
                     | Group( QuotedString('"', "\\") )
                     | Group( QuotedString("'", "\\") )
                     | Group( p_id )
                     )
        p_comp     = ( Literal("==") | Literal("in") | p_name )
        p_selector = ( p_val + p_comp + p_val + StringEnd() )
        try:
            resultlist = p_selector.parseString(selector).asList()
        except ParseException:
            return None
        resultdict = {}
        if resultlist:
            resultdict['val1'] = get_value(resultlist[0])
            resultdict['comp'] = resultlist[1]
            resultdict['val2'] = get_value(resultlist[2])
        return resultdict

    def compile_selector_filter(self, selector):
        """
        Return filter for for testing entities matching a supplied selector.

        Returns None if no selection is performed; i.e. all possible entities are selected.

        Selector formats: see `parse_selector` above.

        This function returns a filter function compiled from the supplied selector.
        """
        def get_entity(field_id):
            "Get field from entity tested by filter"
            def get_entity_f(e, c):
                return e.get(field_id, None)
            return get_entity_f
        #
        def get_context(name, field_id):
            "Get field from named value in current display context"
            def get_context_f(e, c):
                if name in c and c[name]:
                    return c[name].get(field_id, None)
                return None
            return get_context_f
        #
        def get_literal(value):
            "Get literal value specified directly in selector string"
            def get_literal_f(e, c):
                return value
            return get_literal_f
        #
        def get_val_f(selval):
            if selval['type'] == "entity":
                return get_entity(selval['field_id'])
            elif selval['type'] == "context":
                return get_context(selval['name'], selval['field_id'])
            elif selval['type'] == "literal":
                return get_literal(selval['value'])
            else:
                raise ValueError("Unrecognized value type from selector (%s)"%selval['type'])
                assert False, "Unrecognized value type from selector"
        #
        def match_eq(v1f, v2f):
            def match_eq_f(e, c):
                return v1f(e, c) == v2f(e, c)
            return match_eq_f
        #
        def match_in(v1f, v2f):
            def match_in_f(e, c):
                v1 = v1f(e, c)
                if not v1: return True
                v2 = v2f(e, c)
                if isinstance(v2, list):
                    return v1 in v2
                return v1 == v2
            return match_in_f
        #
        def match_subtype(v1f, v2f):
            def match_subtype_f(e, c):
                return self._fieldcomp.subtype(v1f(e, c), v2f(e, c))
            return match_subtype_f
        #
        if selector in {None, "", "ALL"}:
            return None
        sel = self.parse_selector(selector)
        if not sel:
            raise ValueError("Unrecognized selector syntax (%s)"%selector)
        v1f = get_val_f(sel['val1'])
        v2f = get_val_f(sel['val2'])
        if sel['comp'] == "==":
            return match_eq(v1f, v2f)
        if sel['comp'] == "in":
            return match_in(v1f, v2f)
        if sel['comp'] == "subtype":
            return match_subtype(v1f, v2f)
        # Drop through: raise error
        raise ValueError("Unrecognized entity selector (%s)"%selector)

#   -------------------------------------------------------------------
#   FieldComparison
#   -------------------------------------------------------------------

class FieldComparison(object):
    """
    Logic for comparing fields using additional context information not available
    directly to 'EntitySelector'
    """

    def __init__(self, coll):
        super(FieldComparison, self).__init__()
        self._coll     = coll
        self._site     = coll.get_site()
        return

    def get_uri_type_info(self, type_uri, include_alt=True):
        """
        Return typeinfo corresponding to the supplied type URI
        """
        t     = self._coll.get_uri_type(type_uri)
        return t and EntityTypeInfo(self._site, self._coll, t.get_id())

    def subtype(self, type1_uri, type2_uri):
        """
        Returns True if the first type is a subtype of the second type, where both
        types are supplied as type URIs.  Returns True if both URIs are the same.

        If type1_uri is not specified, assume no restriction.

        If type2_uri is not specified, assume it does not satisfy the restriction.
        """
        # log.info("FieldComparison.subtype(%s, %s)"%(type1_uri, type2_uri))
        if not type2_uri or (type1_uri == type2_uri):
            return True
        if not type1_uri:
            return False
        type1_info = self.get_uri_type_info(type1_uri)
        type1_supertype_uris = (type1_info and type1_info.get_all_type_uris()) or []
        # log.info("FieldComparison.subtype: type1_uris (supertypes) %r"%(type1_uris,))
        return type2_uri in type1_supertype_uris

if __name__ == "__main__":
    import doctest
    doctest.testmod()

# End.



"""
am_errors.py - status codes returns by annalist_manager
"""

from __future__ import print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2013-2014, Graham Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

class Annalist_Manager_Error(Exception):
    """
    annalist-manager error
    """
    def __init__(self, errno=None, value=None, msg="Annalist manager error"):
        self._msg   = msg
        self._errno = errno
        self._value = value
        return

    def __str__(self):
        txt = self._msg
        if self._value: txt += ": "+repr(self._value)
        return txt

    def __repr__(self):
        return ( "Annalist_Manager_Error(%s, %s, value=%s)"%
                 (repr(self._msg), repr(self._value)))


# Status return codes

AM_SUCCESS         = 0      # Success
AM_BADCMD          = 2      # Command error
AM_EXISTS          = 5      # directory already exists
AM_NOTEXISTS       = 6      # Directory does not exist
AM_NOSETTINGS      = 7      # No configuration settings found (e.g. personal, shared, devel, etc.)
AM_UNEXPECTEDARGS  = 8      # Unexpected arguments supplied
AM_NOUSERPASS      = 9      # No username or password for createuser or creaeadminuser
AM_MISSINGEMAIL    = 10     # No email address for createuser or creaeadminuser
AM_UNKNOWNCMD      = 11     # Unknown command name for help
AM_USEREXISTS      = 12     # Username for creation already exists
AM_USERNOTEXISTS   = 13     # Username for deletion does not exist

# End.


"""
Tests for file upload field; view as hyperlink.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2015, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import sys
import os
import unittest
import re
from collections import OrderedDict

import logging
log = logging.getLogger(__name__)

from annalist.views.fields.render_file_upload import (
    get_file_upload_renderer, 
    FileUploadValueMapper
    )

from annalist.tests.field_rendering_support import FieldRendererTestSupport

class FileUploadRenderingTest(FieldRendererTestSupport):

    def setUp(self):
        return

    def tearDown(self):
        return

    # Rendering test

    def test_RenderFileUploadValue(self):

        def expect_render(filename, labeltext):
            # render_view = '''<a href="%s" target="_blank">%s</a>'''%(filename, labeltext)
            # render_edit = (
            #     '''<input type="file" name="repeat_prefix_test_field" '''+
            #            '''placeholder="(test placeholder)" '''+
            #            '''value="%s" /> '''
            #     )%filename
            render_view = (
                '''Uploaded file <a href="%s" target="_blank">%s</a>'''
                )%(filename, labeltext)
            render_edit = (
                '''<input type="file" name="repeat_prefix_test_field" '''+
                       '''placeholder="(test placeholder)" '''+
                       '''value="%s" /> '''+
                '''Previously uploaded: %s '''
                )%(filename, labeltext)
            return {'view': render_view, 'edit': render_edit}

        def upload_file_value(file):
            return {'resource_name': file, 'uploaded_file': "uploaded.ext"}

        test_values = (
            [ (upload_file_value("testfile.ext"),   "uploaded.ext")
            ])
        test_value_context_renders = (
            [ (self._make_test_context(filename),  expect_render(filename['resource_name'], labeltext))
                for filename, labeltext in test_values
            ])
        renderer = get_file_upload_renderer()

        for render_context, expect_render in test_value_context_renders:
            # print repr(render_context['field']['field_value'])
            self._check_value_renderer_results(
                renderer,
                context=render_context,
                expect_rendered_view=expect_render['view'],
                expect_rendered_edit=expect_render['edit'],
                collapse_whitespace=True
                )
        return

    def test_DecodeFileUploadValue(self):
        test_decode_values = (
            { None:                 ""
            , "testfile.ext":       "testfile.ext"
            })
        for valtext, expect_valdata in test_decode_values.items():
            valdata = FileUploadValueMapper.decode(valtext)
            self.assertEqual(
                valdata, expect_valdata, 
                "Value decode(%s) = %r, expect %r"%(valtext, valdata, expect_valdata)
                )
        return

# End.

if __name__ == "__main__":
    # import django
    # django.setup()  # Needed for template loader
    # Runtests in this module
    # runner = unittest.TextTestRunner(verbosity=2)
    # tests = unittest.TestSuite()
    # tests  = getSuite(select=sel)
    # if tests: runner.run(tests)
    unittest.main()



"""
Tests for EntityData default editing view
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                    import settings
from django.contrib.auth.models     import User
from django.test                    import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions
from django.test.client             import Client

from annalist.models.site           import Site
from annalist.models.collection     import Collection
from annalist.models.recordtypedata import RecordTypeData
from annalist.models.entitydata     import EntityData

from annalist.views.entitydelete    import EntityDataDeleteConfirmedView

from tests                          import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from tests                          import init_annalist_test_site, resetSitedata
from AnnalistTestCase               import AnnalistTestCase
from entity_testutils               import create_test_user
from entity_testentitydata          import (
    entity_url, entitydata_edit_url, entitydata_delete_confirm_url,
    entitydata_list_all_url, entitydata_list_type_url,
    entitydata_value_keys, entitydata_create_values, entitydata_values, 
    entitydata_delete_confirm_form_data,
    )

class ConfirmEntityDataDeleteTests(AnnalistTestCase):
    """
    Tests for record type deletion on response to confirmation form
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection(self.testsite, "testcoll")
        self.testdata = RecordTypeData(self.testcoll, "testtype")
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    def test_CollectionActionViewTest(self):
        self.assertEqual(EntityDataDeleteConfirmedView.__name__, "EntityDataDeleteConfirmedView", "Check EntityDataDeleteConfirmedView class name")
        return

    # NOTE:  this logic only tests the entity deletion completion code.
    # It is assumed that the response to requesting entity deletion is checked elsewhere.
    def test_post_confirmed_remove_entity(self):
        t = EntityData.create(self.testdata, "deleteentity", entitydata_create_values("deleteentity"))
        self.assertTrue(EntityData.exists(self.testdata, "deleteentity"))
        # Submit positive confirmation
        u = entitydata_delete_confirm_url("testcoll", "testtype")
        f = entitydata_delete_confirm_form_data("deleteentity")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,     302)
        self.assertEqual(r.reason_phrase,   "FOUND")
        self.assertEqual(r.content,         "")
        self.assertMatch(r['location'],    
            "^"+TestHostUri+
            entitydata_list_all_url("testcoll")
            )
        self.assertMatch(r['location'],    
            r"info_head=.*&info_message=.*deleteentity.*testcoll.*$"
            )
        self.assertNotIn("search=testcoll", r['location'])
        # Confirm deletion
        self.assertFalse(EntityData.exists(self.testcoll, "deleteentity"))
        return

    def test_post_confirmed_remove_entity_from_search(self):
        t = EntityData.create(self.testdata, "deleteentity", entitydata_create_values("deleteentity"))
        self.assertTrue(EntityData.exists(self.testdata, "deleteentity"))
        # Submit positive confirmation
        u = entitydata_delete_confirm_url("testcoll", "testtype")
        f = entitydata_delete_confirm_form_data("deleteentity", search="testcoll")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,     302)
        self.assertEqual(r.reason_phrase,   "FOUND")
        self.assertEqual(r.content,         "")
        self.assertMatch(r['location'],    
            "^"+TestHostUri+
            entitydata_list_all_url("testcoll")
            )
        self.assertMatch(r['location'],    
            r"info_head=.*&info_message=.*deleteentity.*testcoll.*$"
            )
        self.assertIn("search=testcoll", r['location'])
        # Confirm deletion
        self.assertFalse(EntityData.exists(self.testcoll, "deleteentity"))
        return

# End.
